#include <stdio.h>
#include <string.h>
#include "uforth.h"

void itoa(int16_t value, char* str, int16_t base);

struct dict ramdict = INITIAL_DICT;
struct dict *dict = &ramdict;

uint8_t rxc(void) {
  return getchar();
}

void rxgetline(char* str) {
  gets(str);
}

void txc(uint8_t c) {
  putchar(c);
}

void txs(char* s, int cnt) {
  while(cnt--) txc(*s++);
}
#define txs0(s) txs(s,strlen(s))

void txn(uint16_t n) {
  static char num[10];
  itoa(n, num, 10);
  txs(num,strlen(num));
}

void stats() { 
  txs0("\tDict: ");
  txn(dict->here);
  txs0(" cells used out of ");
  txn(MAX_DICT_CELLS);
  txs0(" (");
  txn(((uint32_t)dict->here*100)/MAX_DICT_CELLS);
  txs0("% used)\r\n");
}

void interpret_from(FILE *fp);

uforth_stat c_handle(void) {
  CELL r1 = dpop();
  char *str;
  FILE *fp;
  char buf[80];

  switch(r1) {
  case 1:			/* dot */
    if (iram.didx > -1) {
      r1 = dpop();
      txn(r1); txc(' ');
    } else txs0("EMPTY\r\n");
    break;
  case 2:			/* emit */
    txc(dpop()&0xff);
    break;
  case 3:			/* key */
    dpush((CELL)rxc());
    break;
  case 4:			/* stats */
    stats();
    break;
  case 5:			/* type */
    r1 = dpop();
    str = (char*)&(dict->d[dpop()]);
    txs(str,r1);
    break;
  case 6:			/* .s */
    r1 = iram.didx;
    while (r1 >= 0) {
      txn(iram.ds[r1]); txs0("\r\n");
      --r1;
    }
    break;
  case 7:
    {
      uforth_next_word();
      strncpy(buf,iram.currword, iram.currwordlen);
      fp = fopen(buf, "w+");
      fwrite(dict,(sizeof(CELL)*3) + (dict->here)*sizeof(CELL) ,1,fp);
      fclose(fp);
    }
    break;
  case 8:
    {
      uforth_next_word();
      strncpy(buf,iram.currword, iram.currwordlen);
      fp = fopen(buf, "r");
      fread(dict,sizeof(CELL),sizeof(struct dict),fp);
      fclose(fp);
      uforth_abort();
      stats();
    }
    break;
  case 9:
    {
      uforth_next_word();
      strncpy(buf,iram.currword, iram.currwordlen);
      fp = fopen(buf, "r");
      interpret_from(fp);
      fclose(fp);
      stats();
    }  
    break;
  default:
    return app_handle_cmds(r1);
  }
  return OK;
}

void create_c_handle_words() {
  uforth_interpret(": . 1 cf ;");
  uforth_interpret(": emit 2 cf ;");
  uforth_interpret(": key 3 cf ;");
  uforth_interpret(": mem 4 cf ;");
  uforth_interpret(": type 5 cf ;");
  uforth_interpret(": .s 6 cf ;");
  uforth_interpret(": save-image 7 cf ;");
  uforth_interpret(": load-image 8 cf ;");
  uforth_interpret(": load-file 9 cf ;");

  app_create_cmds();
}

static char line[128];
void interpret_from(FILE *fp) {
  int16_t lineno = 0;
  while (!feof(stdin)) {
    ++lineno;
    if (fp == stdin) txs0(" ok\r\n");
    if (fp == stdin)
      rxgetline(line);
    else
      if (fgets(line,128,fp) == NULL) break;
    if (line[0] == '#' || line[0] == '\n' || line[0] == '\0') continue;

    switch(uforth_interpret(line)) {
    case E_NOT_COMPILING:
      txs0("Compile only word!");
    case E_NOT_A_WORD:
      txs0("Huh? >>> ");
      txs(iram.currword,iram.currwordlen);
      txs0(" <<< ");
      txs0(iram.currword + iram.currwordlen);
      txs0("\r\n");
      break;
    case E_ABORT:
      txs0("Abort!\n");
      break;
    case E_STACK_UNDERFLOW:
      txs0("Stack underflow!\n");
      break;
    case E_DSTACK_OVERFLOW:
      txs0("Stack overflow!\n");
      break;
    case E_RSTACK_OVERFLOW:
      txs0("Return Stack overflow!\n");
      break;
    case E_MISSING_LOOP:
      txs0("Missing loop terminator!\n");
      break;
    case E_MISSING_THEN:
      txs0("Missing then!\n");
      break;
    case OK:
      break;
    default:
      txs0("Ugh\n");
      break;
    }
  }
}

int main(int argc, char* argv[]) {
  uforth_init();
  app_init();

  create_c_handle_words();

#ifdef BOOTSTRAP
  {
    FILE *fp;
    printf("BOOTSTRAP Version\n");
    printf("   Loading init.f\n");
    fp = fopen("init.f", "r");
    if (fp != NULL) 
      interpret_from(fp);
    fclose(fp);
  }
#endif

  stats();
  interpret_from(stdin);

  return 0;
}

